/*
 * kgraft_patch_bsc1027575
 *
 * Fix for CVE-2017-2636, bsc#1027575
 *
 *  Upstream commit:
 *  82f2341c94d270421f383641b7cd670e474db56b ("tty: n_hdlc: get rid of racy n_hdlc.tbuf")
 *
 *  SLE12-SP2 commit:
 *  cc5d011563ab92932e49d8bd5cde7b0c65b88adb
 *
 *  SLE12-SP1) commit:
 *  8c968fab953fcc8fda4059a76fbddde024e75dde
 *
 *  SLE12 commit
 *  a71c1d8db7e1ce361c91e20aff041fc2c7a402aa
 *
 *  Copyright (c) 2017 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/printk.h>
#include <linux/bitops.h>
#include <linux/tty.h>

#if IS_ENABLED(CONFIG_N_HDLC)

#if !IS_MODULE(CONFIG_N_HDLC)
#error "KGR patch supports only CONFIG_N_HDLC=m."
#endif

#define KGR_PATCHED_MODULE "n_hdlc"

struct n_hdlc_buf {
	struct n_hdlc_buf *link;
	int		  count;
	char		  buf[1];
};

struct n_hdlc_buf_list {
	struct n_hdlc_buf *head;
	struct n_hdlc_buf *tail;
	int		  count;
	spinlock_t	  spinlock;
};

struct n_hdlc {
	int			magic;
	__u32			flags;
	struct tty_struct	*tty;
	struct tty_struct	*backup_tty;
	int			tbusy;
	int			woke_up;
	struct n_hdlc_buf	*tbuf;
	struct n_hdlc_buf_list	tx_buf_list;
	struct n_hdlc_buf_list	rx_buf_list;
	struct n_hdlc_buf_list	tx_free_buf_list;
	struct n_hdlc_buf_list	rx_free_buf_list;
};

#define DEBUG_LEVEL_INFO	1

static void (*kgr_n_hdlc_buf_put)(struct n_hdlc_buf_list *list,
				  struct n_hdlc_buf *buf);
static struct n_hdlc_buf* (*kgr_n_hdlc_buf_get)(struct n_hdlc_buf_list *list);
static int *kgr_debuglevel;

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "n_hdlc:n_hdlc_buf_put", (void *)&kgr_n_hdlc_buf_put},
	{ "n_hdlc:n_hdlc_buf_get", (void *)&kgr_n_hdlc_buf_get},
	{ "n_hdlc:debuglevel", (void *)&kgr_debuglevel},
};


/*
 * Core of the fix below: rather than storing declined buffers at
 * ->tbuf, return them to the *head* of the TX list. Differs from
 * upstream fix in that it's implemented in terms of the custom,
 * hand-made list implementation: this avoids changing data
 * structures.
 */
static void kgr_n_hdlc_buf_return(struct n_hdlc_buf_list *buf_list,
				  struct n_hdlc_buf *buf)
{
	unsigned long flags;

	spin_lock_irqsave(&buf_list->spinlock, flags);

	buf->link = buf_list->head;
	if (!buf_list->head)
		buf_list->tail = buf;
	buf_list->head = buf;
	buf_list->count++;

	spin_unlock_irqrestore(&buf_list->spinlock,flags);
}

/* patched */
void kgr_n_hdlc_send_frames(struct n_hdlc *n_hdlc, struct tty_struct *tty)
{
	register int actual;
	unsigned long flags;
	struct n_hdlc_buf *tbuf;

	if (*kgr_debuglevel >= DEBUG_LEVEL_INFO)
		printk("%s(%d)n_hdlc_send_frames() called\n",__FILE__,__LINE__);
 check_again:

	spin_lock_irqsave(&n_hdlc->tx_buf_list.spinlock, flags);
	if (n_hdlc->tbusy) {
		n_hdlc->woke_up = 1;
		spin_unlock_irqrestore(&n_hdlc->tx_buf_list.spinlock, flags);
		return;
	}
	n_hdlc->tbusy = 1;
	n_hdlc->woke_up = 0;
	/*
	 * Fix CVE-2017-2636
	 *  -1 line
	 */

	/* get current transmit buffer or get new transmit */
	/* buffer from list of pending transmit buffers */

	tbuf = n_hdlc->tbuf;
	/*
	 * Fix CVE-2017-2636
	 *  +2 lines
	 */
	n_hdlc->tbuf = NULL;
	spin_unlock_irqrestore(&n_hdlc->tx_buf_list.spinlock, flags);
	if (!tbuf)
		tbuf = kgr_n_hdlc_buf_get(&n_hdlc->tx_buf_list);

	while (tbuf) {
		if (*kgr_debuglevel >= DEBUG_LEVEL_INFO)
			printk("%s(%d)sending frame %p, count=%d\n",
				__FILE__,__LINE__,tbuf,tbuf->count);

		/* Send the next block of data to device */
		set_bit(TTY_DO_WRITE_WAKEUP, &tty->flags);
		actual = tty->ops->write(tty, tbuf->buf, tbuf->count);

		/* rollback was possible and has been done */
		if (actual == -ERESTARTSYS) {
			/*
			 * Fix CVE-2017-2636
			 *  -1 line, +1 line
			 */
			kgr_n_hdlc_buf_return(&n_hdlc->tx_buf_list, tbuf);
			break;
		}
		/* if transmit error, throw frame away by */
		/* pretending it was accepted by driver */
		if (actual < 0)
			actual = tbuf->count;

		if (actual == tbuf->count) {
			if (*kgr_debuglevel >= DEBUG_LEVEL_INFO)
				printk("%s(%d)frame %p completed\n",
					__FILE__,__LINE__,tbuf);

			/* free current transmit buffer */
			kgr_n_hdlc_buf_put(&n_hdlc->tx_free_buf_list, tbuf);

			/*
			 * Fix CVE-2017-2636
			 *  -2 lines
			 */

			/* wait up sleeping writers */
			wake_up_interruptible(&tty->write_wait);

			/* get next pending transmit buffer */
			tbuf = kgr_n_hdlc_buf_get(&n_hdlc->tx_buf_list);
		} else {
			if (*kgr_debuglevel >= DEBUG_LEVEL_INFO)
				printk("%s(%d)frame %p pending\n",
					__FILE__,__LINE__,tbuf);

			/* buffer not accepted by driver */
			/* set this buffer as pending buffer */
			/*
			 * Fix CVE-2017-2636
			 *  -1 line, +1 line
			 */
			kgr_n_hdlc_buf_return(&n_hdlc->tx_buf_list, tbuf);
			break;
		}
	}

	if (!tbuf)
		clear_bit(TTY_DO_WRITE_WAKEUP, &tty->flags);

	/* Clear the re-entry flag */
	spin_lock_irqsave(&n_hdlc->tx_buf_list.spinlock, flags);
	n_hdlc->tbusy = 0;
	spin_unlock_irqrestore(&n_hdlc->tx_buf_list.spinlock, flags);

	if (n_hdlc->woke_up)
	  goto check_again;

	if (*kgr_debuglevel >= DEBUG_LEVEL_INFO)
		printk("%s(%d)n_hdlc_send_frames() exit\n",__FILE__,__LINE__);

}	/* end of n_hdlc_send_frames() */


static int kgr_patch_bsc1027575_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1027575_module_notify(struct notifier_block *nb,
	unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1027575_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1027575_module_nb = {
	.notifier_call = kgr_patch_bsc1027575_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1027575_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1027575_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1027575_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1027575_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1027575_module_nb);
}

#endif /* CONFIG_N_HDLC */
